"""

Separate in 3 parts  (classes)

  raw text --> NER extraction


  NER  --> KGraph training + embedding generation/


  KG Embedding Losses



Reason :
   make the code modular, easy to change when needed.
   make thing independant.



"""
import os
import spacy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import networkx as ntx
from tqdm import tqdm
from typing import Tuple, Dict, Union
from spacy.matcher import Matcher
#from node2vec import Node2Vec as n2v

import torch

# from utilmy.data import dataset_download
from pykeen.triples import TriplesFactory
from pykeen.pipeline import pipeline
from pykeen.models import TransE,ERModel
from torch.optim import Adam
from pykeen.training import SLCWATrainingLoop, LCWATrainingLoop
from pykeen.evaluation import RankBasedEvaluator
from pykeen.nn.representation import LabelBasedTransformerRepresentation

### pip install python-box
from box import Box
from utilmy import log



######################################################################################################
def test_all():
    pass


def test1(dirin='final_dataset_clean_v2 .tsv'):

    """
    Doc::

        cd utilmy/nlp/tttorch/kgraph
        python knowledge_graph test1 --dirin mydirdata/
    """
    url = 'https://github.com/arita37/data/raw/main/kgraph_pykeen_small/data_kgraph_pykeen.zip'
    dname = dataset_download(url=url)
    dname = dname.replace("\\", "/")
    path  = os.path.join(dname, 'final_dataset_clean_v2 .tsv')
    df    = pd.read_csv(path, delimiter='\t')

    dname = dname + "/embed/"

    log('##### NER extraction from text ')
    extractor = NERExtractor(df, embeddingFolder=dname, load_spacy=True)
    data_kgf = extractor.extractTriples(sents=-1)
    extractor.export_data(data_kgf)


    log('##### Build Knowledge Graph')
    data_kgf_path = os.path.join(dname, 'data_kgf.tsv')
    data_kgf = knowledge_grapher.load_data(data_kgf_path)
    grapher = knowledge_grapher(data_kgf=data_kgf,embedding_dim=10, load_spacy=True)
    grapher.buildGraph()


    log('##### Build KG Embeddings')
    embedder = KGEmbedder(dname, grapher.graph, embedding_dim=10)
    # If you have the trained model to be saved then pass a non existing dir to load_embeddings()
    embedder.compute_embeddings('none', batch_size=1024)
    embedder.save_embeddings()


    log('##### load KG Embeddings')
    embedder.load_embeddings('none')



def runall(dirin='', dirout='', config=None):

    """  Run all steps to generate dirin
    Doc::

        cd utilmy/nlp/tttorch/kgraph
        python knowledge_graph test1 --dirin mydirdata/

    """
    url = 'https://github.com/arita37/data/raw/main/kgraph_pykeen_small/data_kgraph_pykeen.zip'
    dname = dataset_download(url=url)
    dname = dname.replace("\\", "/")
    path  = os.path.join(dname, 'final_dataset_clean_v2 .tsv')
    df    = pd.read_csv(path, delimiter='\t')

    dname = dname + "/embed/"

    log('##### NER extraction from text ')
    extractor = NERExtractor(df, embeddingFolder=dname, load_spacy=True)
    data_kgf = extractor.extractTriples(sents=-1)
    extractor.export_data(data_kgf)


    log('##### Build Knowledge Graph')
    data_kgf_path = os.path.join(dname, 'data_kgf.tsv')
    data_kgf = knowledge_grapher.load_data(data_kgf_path)
    grapher = knowledge_grapher(data_kgf=data_kgf,embedding_dim=10, load_spacy=True)
    grapher.buildGraph()


    log('##### Build KG Embeddings')
    embedder = KGEmbedder(dname, grapher.graph, embedding_dim=10)
    # If you have the trained model to be saved then pass a non existing dir to load_embeddings()
    embedder.compute_embeddings('none', batch_size=1024)
    embedder.save_embeddings()


    log('##### load KG Embeddings')
    embedder.load_embeddings('none')



######################################################################################################
class knowledge_grapher():
    def __init__(self, data_kgf, embedding_dim:int=14, load_spacy:bool=False,
                dirin:str="./mydatain/",
                 dirout:str="./mydataout/",



                 ) -> None:
        self.data_kgf = data_kgf
        self.embedding_dim = embedding_dim

    def buildGraph(self, relation = None)->None:
        if relation:
            self.graph = ntx.from_pandas_edgelist(self.data_kgf[self.data_kgf['edge']==relation], "source", "target",
                         edge_attr=True, create_using=ntx.MultiDiGraph())
        else:
            self.graph = ntx.from_pandas_edgelist(self.data_kgf, "source", "target",
                         edge_attr=True, create_using=ntx.MultiDiGraph())

    def plot_graph(self, plotFolder)->None:
        plt.figure(figsize=(14, 14))
        posn = ntx.spring_layout(self.graph)
        ntx.draw(self.graph, with_labels=True, node_color='green', edge_cmap=plt.cm.Blues, pos = posn)
        plt.savefig(os.path.join(plotFolder,'graphPlot.jpg'))
        plt.close()

    def compute_centrality(self,)->None:
        self.centrality_dict = ntx.degree_centrality(self.graph)
        self.in_centrality_dict = ntx.in_degree_centrality(self.graph)
        self.out_centrality_dict = ntx.out_degree_centrality(self.graph)
        # self.eigenvector_centrality_dict = ntx.katz_centrality(self.graph)

    @staticmethod
    def load_data(path)->pd.DataFrame:
        return pd.read_csv(path, delimiter='\t')
        # self.buildGraph(data_kgf)

    def get_centers(self, max_centers:int=5)->None:

        sorted_dict = sorted(self.centrality_dict.items(), key=lambda x: x[1])[::-1]
        in_sorted_dict = sorted(self.in_centrality_dict.items(), key=lambda x: x[1])[::-1]
        out_sorted_dict = sorted(self.out_centrality_dict.items(), key=lambda x: x[1])[::-1]

        degree_centers = sorted_dict[:max_centers]

        in_degree_centers = in_sorted_dict[:max_centers]
        out_degree_centers = out_sorted_dict[:max_centers]

        degree_adjacency = {u:self.graph[u] for u,_ in degree_centers}

        in_degree_adjacency = {u:self.graph[u] for u,_ in in_degree_centers}
        out_degree_adjacency = {u:self.graph[u] for u,_ in out_degree_centers}

        self.center_dict = {'degree':{'centers':degree_centers, 'adjacency': degree_adjacency},
                            'in_degree':{'centers':in_degree_centers, 'adjacency': in_degree_adjacency},
                            'out_degree':{'centers':out_degree_centers, 'adjacency':out_degree_adjacency},}

    def map_centers_anchors(self,embedding_df:pd.DataFrame, _type:str):

       self.embedding_df = embedding_df
       _aux = self.center_dict[_type]
       centers = _aux['centers']
       adjacency = _aux['adjacency']

       self.mean_anchor_dict = {}
       for center, _ in centers:
            center_embedding = self.embedding_df[str(center)]
            num_embeddings = len(list(adjacency[center].keys()))
            adjacency_embeddings = np.ndarray((num_embeddings, self.embedding_dim))
            for i, (node, adj_dict) in enumerate(adjacency[center].items()):
                adjacency_embeddings[i,:] = self.embedding_df[str(node)]
            self.mean_anchor_dict[center] = {'center': center_embedding.values, 'anchor':adjacency_embeddings.mean(axis = 0)}



class NERExtractor:

    def __init__(self, data:pd.DataFrame, embeddingFolder:str, load_spacy=True,
                 dirin:str="./mydatain/",
                 dirout:str="./mydataout/",


                 ):

        if load_spacy:
            self.nlp = spacy.load("ro_core_news_sm")
        self.data = data
        self.embeddingFolder = embeddingFolder

    def extract_entities(self, sents)->pd.DataFrame:
        # chunk one
        enti_one = ""
        enti_two = ""
        dep_prev_token = "" # dependency tag of previous token in sentence
        txt_prev_token = "" # previous token in sentence
        prefix = ""
        modifier = ""
        for tokn in self.nlp(sents):
            # chunk two
            ## move to next token if token is punctuation
            if tokn.dep_ != "punct":
                #  check if token is compound word or not
                if tokn.dep_ == "compound":
                    prefix = tokn.text
                    # add the current word to it if the previous word is 'compound’
                    if dep_prev_token == "compound":
                        prefix = txt_prev_token + " "+ tokn.text
                # verify if token is modifier or not
                if tokn.dep_.endswith("mod") == True:
                    modifier = tokn.text
                    # add it to the current word if the previous word is 'compound'
                    if dep_prev_token == "compound":
                        modifier = txt_prev_token + " "+ tokn.text
                # chunk3
                if tokn.dep_.find("subj") == True:
                    enti_one = modifier +" "+ prefix + " "+ tokn.text
                    prefix = ""
                    modifier = ""
                    dep_prev_token = ""
                    txt_prev_token = ""
                # chunk4
                if tokn.dep_.find("obj") == True:
                    enti_two = modifier +" "+ prefix +" "+ tokn.text
                # chunk 5
                # update variable
                dep_prev_token = tokn.dep_
                txt_prev_token = tokn.text
        return [enti_one.strip(), enti_two.strip()]

    def obtain_relation(self,sent):
        doc = self.nlp(sent)
        matcher = Matcher(self.nlp.vocab)
        pattern = [{'DEP':'ROOT'},
                {'DEP':'prep','OP':"?"},
                {'DEP':'agent','OP':"?"},
                {'POS':'ADJ','OP':"?"}]
        matcher.add(key="matching_1", patterns = [pattern])
        matcher = matcher(doc)
        h = len(matcher) - 1
        try:
            assert matcher
            span = doc[matcher[h][1]:matcher[h][2]]
            return span.text
        except AssertionError:
            print('No match found for this entry!')
            return None

    def extractTriples(self, max_text:int) -> pd.DataFrame:

        pairs_of_entities = [self.extract_entities(i) for i in tqdm(self.data['paragraph'][:max_text])]
        relations = [self.obtain_relation(j) for j in tqdm(self.data['paragraph'][:max_text])]
        indexes = [x for x, z in enumerate(relations) if z is not None]
        relations = [x for x in relations if x is not None]

        # subject extraction
        source = [j[0] for j in pairs_of_entities]
        source = [source[i] for i in indexes]

        #object extraction
        target = [k[1] for k in pairs_of_entities]
        target = [target[i] for i in indexes]

        return pd.DataFrame({'source':source, 'target':target, 'edge':relations})

    def export_data(self, data_kgf:pd.DataFrame)->Tuple[pd.DataFrame]:

        from utilmy import pd_to_file
        train_df, val_df, test_df = dataset_traintest_split(data_kgf, train_ratio=0.6, val_ratio=0.2)

        # SAMPLES = len(data_kgf.index)
        # TRAIN_SPLIT = int(0.5 * SAMPLES)
        # TEST_SPLIT = int(0.3 * SAMPLES)
        # VALIDATION_SPLIT = int(0.2 * SAMPLES)
        #
        # train_indexes = np.random.randint(low = 0, high = len(data_kgf.index), size=TRAIN_SPLIT)
        # test_indexes = np.random.randint(low = 0, high = len(data_kgf.index), size=TEST_SPLIT)
        # validation_indexes = np.random.randint(low = 0, high = len(data_kgf.index), size=VALIDATION_SPLIT)
        #
        # train_df = data_kgf.iloc[train_indexes]
        # test_df = data_kgf.iloc[test_indexes]
        # val_df = data_kgf.iloc[validation_indexes]


        train_df.to_csv(os.path.join(self.embeddingFolder,'train_data.tsv'), sep="\t")
        test_df.to_csv(os.path.join(self.embeddingFolder,'test_data.tsv'), sep="\t")
        val_df.to_csv(os.path.join(self.embeddingFolder,'validation_data.tsv'), sep="\t")
        data_kgf.to_csv(os.path.join(self.embeddingFolder,'data_kgf.tsv'), sep="\t")

        return train_df, test_df, val_df



class KGEmbedder:
    def __init__(self, dataFolder, graph:ntx.MultiDiGraph, embedding_dim:int,
                 dirin:str="./mydatain/",
                 dirout:str="./mydataout/",

                 )->None:

        self.dataFolder = dataFolder
        self.graph = graph
        self.embedding_dim = embedding_dim

        train_path =os.path.join(dataFolder,'train_data.tsv')
        test_path =os.path.join(dataFolder,'test_data.tsv')
        val_path =os.path.join(dataFolder,'validation_data.tsv')
        data_path = os.path.join(dataFolder,'data_kgf.tsv')


        self.training = TriplesFactory.from_path(train_path)

        self.testing = TriplesFactory.from_path(test_path,
                                            entity_to_id=self.training.entity_to_id,
                                            relation_to_id=self.training.relation_to_id)

        self.validation = TriplesFactory.from_path(val_path,
                                            entity_to_id=self.training.entity_to_id,
                                            relation_to_id=self.training.relation_to_id)


    def set_up_embeddings(self,):
        # entity_representations = LabelBasedTransformerRepresentation.from_triples_factory(training)

        if os.path.exists(os.path.join(self.dataFolder, 'trained_model.pkl')):
            self.model = torch.load(os.path.join(self.dataFolder, 'trained_model.pkl'))
            self.trained = True
        else:
            self.model = ERModel(triples_factory=self.training,
                                 interaction='distmult',
                               # entity_representations=entity_representations
                               entity_representations_kwargs = dict(embedding_dim=self.embedding_dim, dropout=0.1),
                               relation_representations_kwargs = dict(embedding_dim=self.embedding_dim, dropout=0.1)
                               )

            self.optimizer = Adam(params=self.model.get_grad_params())

            self.training_loop = LCWATrainingLoop(
                model=self.model,
                triples_factory=self.training,
                optimizer=self.optimizer,
            )
            self.trained = False

    def compute_embeddings(self, path_to_embeddings, batch_size=1024):

        self.set_up_embeddings()
        if not self.trained:
            losses = self.training_loop.train(
                                triples_factory=self.training,
                                num_epochs=10,
                                checkpoint_name='myCheckpoint.pt',
                                checkpoint_frequency=5,
                                batch_size=256,
                                )
            torch.save(self.model, os.path.join(self.dataFolder, 'trained_model.pkl'))
        else:
            losses = None

        # Pick an evaluator
        evaluator = RankBasedEvaluator()
        # Get triples to test

        mapped_triples = self.testing.mapped_triples
        # Evaluate
        results = evaluator.evaluate(
            model=self.model,
            mapped_triples=mapped_triples,
            batch_size=batch_size,
            additional_filter_triples=[
                self.training.mapped_triples,
                self.validation.mapped_triples,
            ],
        )
        return losses, results


    def load_embeddings(self, path_to_embeddings:str):

        if os.path.exists(path_to_embeddings):
           self.embedding_df = pd.read_csv(path_to_embeddings)
           return None, None
        #else:
        #    return self.compute_embeddings(path_to_embeddings, batch_size=1024)

    def save_embeddings(self,):

        entities = tuple(self.graph.nodes.values())
        tripleFactory = self.training

        entities_to_ids:Dict = tripleFactory.entity_id_to_label
        relation_to_ids:Dict = tripleFactory.relation_id_to_label

        # TransE model has only one embedding per entity/relation
        entity_embeddings = self.model.entity_representations[0]
        relation_embeddings = self.model.relation_representations[0]

        self.relation_dict = get_embeddings(relation_to_ids, relation_embeddings)
        self.entity_dict = get_embeddings(entities_to_ids, entity_embeddings)
        df_entities = embeddingsToDF(self.entity_dict, 'entity')
        df_relation = embeddingsToDF(self.relation_dict, 'relation')

        df_entities.to_parquet(os.path.join(self.dataFolder, 'entityEmbeddings.parquet'))
        df_relation.to_parquet(os.path.join(self.dataFolder, 'relationEmbeddings.parquet'))



######################################################################################################
def dataset_traintest_split(anyobject, train_ratio=0.6, val_ratio=0.2):
    #### Split anything
    val_ratio = val_ratio + train_ratio
    if isinstance(anyobject, pd.DataFrame):
        df = anyobject
        itrain,ival = int(len(df)* train_ratio), int(len(df)* val_ratio)
        df_train = df.iloc[0:itrain,:]
        df_val   = df.iloc[itrain:ival,:]
        df_test  = df.iloc[ival:,:]
        return df_train, df_val, df_test

    else :  ## if isinstance(anyobject, list):
        df = anyobject
        itrain,ival = int(len(df)* train_ratio), int(len(df)* val_ratio)
        df_train = df[0:itrain]
        df_val   = df[itrain:ival]
        df_test  = df[ival:]
        return df_train, df_val, df_test




def dataset_download(url    = "https://github.com/arita37/data/raw/main/fashion_40ksmall/data_fashion_small.zip",
                     dirout = "./"):
    """ Downloading Dataset from github  and unzip it


    """
    import requests
    from zipfile import ZipFile

    fname = dirout + url.split("/")[-1]
    dname = dirout + "/".join( fname.split(".")[:-1])

    isdir = os.path.isdir(dname)
    if isdir == 0:
        r = requests.get(url)
        open(fname , 'wb').write(r.content)
        flag = os.path.exists(fname)
        if(flag):
            print("Dataset is Downloaded")
            zip_file = ZipFile(fname)
            zip_file.extractall()
        else:
            raise Exception("Dataset is not downloaded")
    else:
        print("dataset is already presented")
    return dname




def get_embeddings(label_to_id:Dict[str, int], embedding):
    aux = {label:{'_id': id_} for id_, label in label_to_id.items()}
    for id_, label in label_to_id.items():
        idx_tensor = torch.tensor([id_]).to(dtype=torch.int64)

        aux[label]['embedding'] = embedding.forward(indices = idx_tensor).detach().numpy()
    return aux



def embeddingsToDF(embeddingDict:Dict[str, Dict[str, Union[int, torch.tensor]]], entityOrRelation:str)->pd.DataFrame:
    aux = []
    for label, dict_ in embeddingDict.items():
        vals = [label, dict_['_id'], dict_['embedding'].flatten()]
        aux.append(vals)
    df = pd.DataFrame(aux, columns=[entityOrRelation, 'id', 'embedding'])
    return df




if __name__=="__main__":
    import fire
    fire.Fire()
    # runall()
